#include "stdafx.h"
#include "ALGORITHM.h"
#include "lrsstatreader.h"
#include <math.h>
#include "Hierarchies.h"
#include "NAVIGATOR.h"


// Used only for the 'Pull' case processor sample code
#include "pullcaseprocessor.h"


/*============================================================================
  File:      IDMAlgorithm.Stubs.cpp
 
  Summary:   IDMAlgorithm stubs to paste into implementation file for class 
			 implementing the new algorithm

  Date:		 September 12, 2003
------------------------------------------------------------------------------
  This file is part of the Microsoft SQL Server Code Samples.
 
  Copyright (C) 2003 Microsoft Corporation.  All rights reserved.
 
This source code is intended only as a supplement to Microsoft
Development Tools and/or on-line documentation.  See these other
materials for detailed information regarding Microsoft code samples.
 
THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.
============================================================================*/

////
//
// IDMAlgorithm interface
//
////


STDMETHODIMP ALGORITHM::Initialize( 
					/* [in] */ IDMContextServices*	in_pExeContext,
					/* [in] */ IDMAttributeSet*		in_pAttributeSet,
					/* [in] */ IDMMarginalStat*		in_pMarginalStat)
{
	// Provide initialization of the algorithm. All allocations are done
	// with the context-services interface.

	// The algorithm will probably want to cache away the attribute-set
	// interface, or alternatively, initialize some other representation
	// for the set of attributes by using this interface.

	// E.g., in the definition of ALGORITHM include the member:
	//
	//	CComPtr<IDMAttributeSet> _spidmattributeset;
	// 
	// and then:

	_spidmattributeset = in_pAttributeSet;

	// The marginal-statistics interface contains marginal counts for
	// each attribute in the training data. These pre-computed counts may
	// be useful to the algorithm - you save a data pass by using this
	// interface instead of scanning the data.
	
	return S_OK;
}

STDMETHODIMP ALGORITHM::InsertCases( 
					/* [in] */ IDMContextServices *in_pExeContext,
					/* [in] */ IDMPushCaseSet *in_pCaseSet,
					/* [in] */ ULONG in_ulNumParameters,
					/* [in] */ WCHAR *in_rgParameterNames[  ],
					/* [in] */ DMVariantPtr in_rgParameterValues[  ])
{
	// This function is the main "train the model" function. 
	// 
	// The last three inputs to the function allow algorithm-specific parameters
	// to be passed in. If the algorithm's behaviour depends on one or more such
	// parameters, you must extract the appropriate values.
	//
	// If you use the static parameter handling functionality, you can
	// extract these parameters via the static DMPARAMHANDLER member of your
	// algorithm class. Suppose, for example, that you have a type-double
	// parameter called "MY_DOUBLE" that was defined using the DECLARE_PARAMETER
	// macro. You can extract the value via:
	HRESULT hr;
	hr = _dmhparamhandler.GetParameterValue(	in_pExeContext, 
										_dblMinDepScore, 
										L"MINIMUM_DEPENDENCY_SCORE",
										in_ulNumParameters,
										in_rgParameterNames,
										in_rgParameterValues);
	RETURN_ON_FAIL(hr);

	hr = _dmhparamhandler.GetParameterValue(	in_pExeContext, 
										_bDisplayCorrelation, 
										L"DISPLAY_CORRELATION",
										in_ulNumParameters,
										in_rgParameterNames,
										in_rgParameterValues);
	RETURN_ON_FAIL(hr);

	DMHALLOC	dmhalloc; // Memory-allocation wrapper
	dmhalloc.SetAllocator(in_pExeContext);	

	LRSSTATREADER lrsstatreader(dmhalloc);
	
	// Initialize the reader with the attributes
	lrsstatreader.Initialize(_spidmattributeset);


// NOTE: Define USE_PULL_CASESET here to enable the Push-to-pull case processing sample code
#ifndef USE_PULL_CASESET
	// Create a case processor
	CComObjectDM<CASEPROCESSOR>*	pcaseprocessor = NULL;
	hr = CComObjectDM<CASEPROCESSOR>::CreateInstance(dmhalloc, &pcaseprocessor);
	RETURN_ON_FAIL(hr);
	


	// Initialize the case processor with our reader
	hr = pcaseprocessor->Initialize(_spidmattributeset, in_pCaseSet, &lrsstatreader);
	RETURN_ON_FAIL(hr);

	// QI the CASEPROCESSOR for IDMCaseProcessor
	CComPtr<IDMCaseProcessor>	spidmcaseprocessor;
	hr = pcaseprocessor->QueryInterface(&spidmcaseprocessor);
	RETURN_ON_FAIL(hr);
	
	// Load in all of the statistics
	hr = in_pCaseSet->StartCases(spidmcaseprocessor, false /*need case id*/);
	RETURN_ON_FAIL(hr);
#else
	PULLCASEPROCESSOR	pullCP;
	hr	=	pullCP.SetAllocator( in_pExeContext );
	RETURN_ON_FAIL(hr);

	hr	=	pullCP.Initialize( _spidmattributeset, in_pCaseSet, &lrsstatreader);
	RETURN_ON_FAIL(hr);

	hr	=	pullCP.ProcessCases(in_pExeContext);
	RETURN_ON_FAIL(hr);
#endif


	// We must use the model-level allocator for _lrpmodel because 
	// this member's data will live past the lifetime of the current
	// call.
	_lrpmodel.SetAllocator(*this);
	hr = _lrpmodel.PopulateModel(lrsstatreader, _dblMinDepScore);

	return hr;
}


STDMETHODIMP ALGORITHM::Predict( 
    /* [in] */ IDMContextServices*	in_pContext,
    /* [in] */ DM_PREDICTION_FLAGS	in_Flags,
    /* [in] */ IDMAttributeGroup*	in_pPredAttGroup,
    /* [in] */ DM_CaseID			in_CaseID,
    /* [in] */ ULONG				in_ulCaseValues,
    /* [in] */ DM_ATTRIBUTE_VALUE*	in_rgValues,
    /* [in] */ ULONG				in_ulMaxPredictions,
    /* [in] */ ULONG				in_ulMaxStates,
    /* [out] */ DM_ATTRIBUTE_STAT**	io_prgPredictions,
    /* [out] */ ULONG*				out_pulPredictions)
{
	// This is the function used to extract predictions from your model. 
	//
	// The output variables (i.e., the variables for which you are being requested
	// to return predictions) are specified in a "group" using IDMAttributeGroup; you 
	// iterate through this group using IDMAttributeGroup::Reset() and 
	// IDMAttributeGroup::Next(). Some of the attributes in this group, however, may
	// not be allowed to be predicted, based on the input DM_PREDICTION_FLAGS in_flags.
	//
	// The known input variables are supplied in a "case" of attribute/value pairs
	// in_rgValues (this array has in_ulCaseValues elements). 
	//
	// ************* IMPORTANT NOTE: 
	// If in_rgValues[i].Attribute is equal to DM_UNSPECIFIED, you are to ignore it in the 
	// input case!!!
	//
	// The output distributions are to be placed in io_prgPredictions: if the contents of
	// this pointer is NULL, you need to allocate the appropriate amount of space for it 
	// using the in_pContext; otherwise, if io_prgPredictions is not NULL, you are guaranteed by
	// the caller that out_pulPredictions has enough space to hold all of the necessary
	// predictions.
	//
	// You should limit the output list to at most in_ulMaxPredictions, where each
	// discrete-variable prediction contains the description for at most 
	// in_ulMaxStates states.
	// 
	// The supplied DM_PREDICTION_FLAGS determine the behaviour of the call:
	//
	//	DMPF_EXCLUDE_INPUT_ATTRIBUTES			Don't predict any attribute that is explicitly
	//											specified in in_rgValues
	//
	//	DMPF_ONLY_INPUT_ATTRIBUTES				Only predict those values that are explicitly
	//											specified in in_rgValues
	//
	//	DMPF_EXCLUDE_MISSING_STATE				Don't return a prediction about the missing
	//											state
	//
	//	DMPF_ATTRIBUTE_SORT_ASCENDING			Sort io_prgPredictions in ascending order, based
	//											on the value of (in_flags | DMPF_ATT_SORT_MASK)
	//
	//	DMPF_STATE_SORT_ASCENDING				Sort the states of each prediction in ascending
	//											order, based on the value of 
	//											(in_flags | DMPF_STATE_SORT_MASK)
	//
	//	DMPF_INCLUDE_STATS						Populate the rgStateStats member for each 
	//											distribution. Similar to the io_prgPredictions passed in,
	//											if rgStateStats is NULL, you need to allocate the
	//											necessary memory; otherwise, you are guaranteed by
	//											the caller that rgStateStats contains enough space
	//											for your target distribution.
	//
	//	DMPF_INCLUDE_NODE_ID					Populate DM_ATTRIBUTE_STAT::strNodeID for each
	//											prediction. This ID is supposed to be "relevant"
	//											to the prediction that was made, and may be
	//											distinct from the "drill-through" id and/or the
	//											"mining model dimension" id. (See GetNodeIDsForCase).
	//
	//	DMPF_INCLUDE_TIME						Not used for this function.
	//
	//	DMPF_ATT_SORT_MASK						Mask for determining how to sort attributes
	//
	//	DMPF_ATT_SORT_EXIST_SUPPORT				Sort attributes by support (i.e. number
	//											of cases in the training data)
	//
	//	DMPF_ATT_SORT_EXIST_PROB				Sort attributes by posterior probability of
	//											non-missing.
	//
	//	DMPF_ATT_SORT_EXIST_ADJUST_PROB			Sort attributes by "adjusted probability". The
	//											adjusted probability is used to implement (e.g.)
	//											a penalty for popular items.
	//
	//	DMPF_STATE_SORT_MASK					Mask for determining how to sort discrete states
	//
	//	DMPF_STATE_SORT_PROB					Sort states by posterior probability
	//
	//	DMPF_STATE_SORT_SUPPORT					Sort states by support
	//
	//	DMPF_STATE_SORT_VARIANCE				Sort states by posterior variance
	//
	//	DMPF_STATE_SORT_ADJUST_PROB				Sort states by "adjusted probability"
	//
	//
	// A note on representing the posterior of a continuous distribution
	// -----------------------------------------------------------------
	// All continuous distributions are necessarily "binary continuous": there's a probability of 
	// missing vs. present, and then a conditional (continuous) distribution given that the attribute 
	// is present. The continuous posterior must is characterized (for the purposes of Predict) by
	// a mean and variance.
	//
	// Assuming the caller has requested Predict() to include statistics, the (DM_STATE_STAT*) 
	// rgStateStats member of each DM_ATTRIBUTE_STAT must be allocated to size = 2; one element 
	// specifies the distribution of the missing state, and the other element specifies both 
	// p(present) and the conditional distribution given that the attribute is present.
	// 
	// Suppose we have the following posterior distribution:
	//
	//
	// p(x)    = {    p(x is missing)                             if x is missing
    //				  p(x not missing) * p(x | x not missing)     if x is not missing
	//		     }
	//
	// and we want to populate a DM_ATTRIBUTE_STAT called dmattributestat. First, we would 
	// allocate two members for dmattributestat.rgStateStats. Then,
	//
	// // Specify p(x is not missing) and p(x | x is not missing) using a DM_STATE_STAT: 
	// // the posterior mean is specified using the Value member in this case
	// dmattributestat.rgStateStats[1].Value = DmstatevalueContinuous(/* posterior mean (given not missing)*/); 
	// dmattributestat.rgStateStats[1].dblVariance = // posterior variance (given not missing)
	// dmattributestat.rgStateStats[1].dblProbability = // p(x is not missing)
	// dmattributestat.rgStateStats[1].dblSupport =		// Number cases where x is not missing
	// dmattributestat.rgStateStats[1].dblProbabilityVariance = 0; // Variance of p(x is not missing). Not used.
	//
	// // Specify p(x is missing): put missing as the Value, and specify the probability of missing.
	// dmattributestat.rgStateStats[0].Value = DmstatevalueContinuous(dblMissing); 
	// dmattributestat.rgStateStats[0].dblProbability = // p(x is missing)
	// dmattributestat.rgStateStats[0].dblSupport =		// Number cases where x is missing
	// dmattributestat.rgStateStats[0].dblVariance = 0; // N/A: just set to 0
	// dmattributestat.rgStateStats[0].dblProbabilityVariance = 0; // Variance of p(x is missing). Not used.
	//

    
	// We need a dense representation of the case to pass into _lrpmodel.ExtractPosterior()
	// Also, we'll set up a vector of bools to indicate whether or not each attribute
	// has been specified as an input or not. 

	VDBL	vdblCaseDense(*this);
	VB		vbInput(*this);

	ULONG cAttribute = _lrpmodel._cAttribute;

	vdblCaseDense.resize(cAttribute, dblMissing);
	vbInput.resize(cAttribute, false);

	// AVPair = attribute/value pair
	for (UINT iAVPair = 0; iAVPair < in_ulCaseValues; iAVPair++)
	{
		DM_ATTRIBUTE_VALUE& dmattributevalue = in_rgValues[iAVPair];
		ULONG iAttribute = dmattributevalue.Attribute;

		// Check if the attribute is "unspecified" and if so, ignore.
		if (iAttribute == DM_UNSPECIFIED)
		{
			continue;
		}
			
		// Note that the translation of the values MUST match that which
		// is done in the CASEPROCESSOR in InsertCases(). Because we 
		// only have continuous variables, this is a non-issue for us, but 
		// you need to pay attention to this for categorical variables.
		// See DataValues.h for details.
		DBL	 dblValue	= ::DblContinuous(dmattributevalue.Value);
		vdblCaseDense[iAttribute]	= dblValue;
		vbInput[iAttribute]			= true;
	}
	
	// Check if we need to resize io_prgPredictions. If so, cycle over the attributes to
	// count.

	CComPtr<IDMMemoryAllocator>	spidmmemoryallocator	= NULL;
	CComPtr<IDMStringHandler>	spidmstringhandler		= NULL;

	if (io_prgPredictions == NULL)
	{
		ATLASSERT(false);
		return E_POINTER;
	}

	if (*io_prgPredictions == NULL)
	{
		// Need to allocate memory
		HRESULT hr	=	in_pContext->GetVectorAlloc(&spidmmemoryallocator);
		RETURN_ON_FAIL(hr);

		// Count the number of outputs
		ULONG	iAttributeOutput	= 0;
		UINT	cAttributePredict	= 0;
		while ((hr = in_pPredAttGroup->Next(&iAttributeOutput)) != S_FALSE)
		{
			RETURN_ON_FAIL(hr);

			if ((in_Flags & DMPF_EXCLUDE_INPUT_ATTRIBUTES) && vbInput[iAttributeOutput])
			{
				continue;
			}
			else if ((in_Flags & DMPF_ONLY_INPUT_ATTRIBUTES) && !vbInput[iAttributeOutput])
			{
				continue;
			}

			cAttributePredict++;
		}

		if (cAttributePredict > in_ulMaxPredictions)
		{
			cAttributePredict = in_ulMaxPredictions;
		}

		hr	=	spidmmemoryallocator->Alloc(cAttributePredict*sizeof(DM_ATTRIBUTE_STAT), (byte**) io_prgPredictions);
		
		RETURN_ON_FAIL(hr);

		for (ULONG iAttribute = 0; iAttribute < cAttributePredict; iAttribute++)
		{
			// Make sure that rgStateStats is NULL so that we properly allocate

			DM_ATTRIBUTE_STAT& dmattributestat = *io_prgPredictions[iAttribute];

			dmattributestat.Attribute							= -1;
			dmattributestat.strNodeID							= (DMStringPtr) NULL;
			dmattributestat.dblMin								= 0.0;
			dmattributestat.dblMax								= 0.0;
			dmattributestat.dblExistingSupport					= 0.0;
			dmattributestat.dblExistingProbability				= 0.0;
			dmattributestat.dblExistingAdjustedProbability		= 0.0;
			dmattributestat.ulStates							= 0;
			dmattributestat.rgStateStats						= NULL;
		}
	}
	
	// Cycle over the set of variables (possibly for the second time) and build the
	// predictions.

	HRESULT hr = in_pPredAttGroup->Reset();
	RETURN_ON_FAIL(hr);
	ULONG	iAttributeOutput	= 0;
	UINT	iPrediction			= 0;

	while ((hr = in_pPredAttGroup->Next(&iAttributeOutput)) != S_FALSE)
	{
		RETURN_ON_FAIL(hr);

		// Check for exclusions using DM_PREDICTION_FLAGS
		if ((in_Flags & DMPF_EXCLUDE_INPUT_ATTRIBUTES) && vbInput[iAttributeOutput])
		{
			continue;
		}
		else if ((in_Flags & DMPF_ONLY_INPUT_ATTRIBUTES) && !vbInput[iAttributeOutput])
		{
			continue;
		}

		// Put together the prediction
		DM_ATTRIBUTE_STAT& dmattributestat = *io_prgPredictions[iPrediction++];
		dmattributestat.Attribute = iAttributeOutput;
		if (in_Flags & DMPF_INCLUDE_NODE_ID)
		{
			// The caller is requesting a node id that corresponds to the
			// prediction. We'll simply use: "PredictionIDX", where X is the
			// attribute number. 
			
			if (spidmstringhandler == NULL)
			{
				hr = in_pContext->GetStringHandler(&spidmstringhandler);

				RETURN_ON_FAIL(hr);
			}
			
			WCHAR szUniqueName[200];
			swprintf_s(szUniqueName, 200, L"PredictionID%d", iAttributeOutput);
			UINT cch = (UINT) wcslen(szUniqueName);
			
			hr = spidmstringhandler->CreateNewStringHandle(&dmattributestat.strNodeID);
			RETURN_ON_FAIL(hr);
			
			hr = spidmstringhandler->CopyBufferToHandle(dmattributestat.strNodeID,
														szUniqueName,
														cch);


			RETURN_ON_FAIL(hr);
		}

		// Min and max values
		dmattributestat.dblMin = _lrpmodel._vdblMin[iAttributeOutput];
		dmattributestat.dblMax = _lrpmodel._vdblMax[iAttributeOutput];

		// No missing values
		dmattributestat.dblExistingSupport		= _lrpmodel._cCase;
		dmattributestat.dblExistingProbability	= 1.0;
    
		// Not implementing an adjusted probability
		dmattributestat.dblExistingAdjustedProbability = 1.0;

		if (in_Flags & DMPF_INCLUDE_STATS)
		{
			// As described above, we have two states: one for missing, which
			// has probability zero in our case, and one for present, which will
			// also describe our continuous distribution.
			dmattributestat.ulStates = 2;
			if (in_ulMaxStates < 2)
			{
				if (in_ulMaxStates == 1)
				{
					dmattributestat.ulStates = 1;
				}
				else
				{
					return E_FAIL;
				}
			}

			
			if (dmattributestat.rgStateStats == NULL)
			{
				// Allocate the space
				if (spidmmemoryallocator == NULL)
				{
					hr	=	in_pContext->GetVectorAlloc(&spidmmemoryallocator);
						RETURN_ON_FAIL(hr);
				}

				hr	=	spidmmemoryallocator->Alloc(dmattributestat.ulStates*sizeof(DM_STATE_STAT), (byte**) &dmattributestat.rgStateStats);
				RETURN_ON_FAIL(hr);
			}
			memset( dmattributestat.rgStateStats, 0, dmattributestat.ulStates*sizeof(DM_STATE_STAT));

			// We know that regardless of the sorting criterion, the present state will always have 
			// higher "score" than the missing state. Thus:
			INT iStateStatPresent = 0;
			INT iStateStatMissing = 1;
			if (in_Flags & DMPF_STATE_SORT_ASCENDING)
			{
				// Caller wants lower-probability states first
				iStateStatPresent = 1;
				iStateStatMissing = 0;
			}

			// Get the posterior
			DBL dblPostMean = 0;
			DBL dblPostSD	= 0;
			hr = _lrpmodel.ExtractPosterior(vdblCaseDense, iAttributeOutput, dblPostMean, dblPostSD);
			RETURN_ON_FAIL(hr);
			if (iStateStatPresent < (INT) dmattributestat.ulStates)
			{
				// p(x)		: Use the posterior mean for the value 
				dmattributestat.rgStateStats[iStateStatPresent].Value		= ::DmstatevalueContinuous(dblPostMean); 
				dmattributestat.rgStateStats[iStateStatPresent].dblVariance = sqrt(dblPostSD);

				// p(present) = 1
				dmattributestat.rgStateStats[iStateStatPresent].dblProbability			= 1.0;
				dmattributestat.rgStateStats[iStateStatPresent].dblSupport				= _lrpmodel._cCase;
				dmattributestat.rgStateStats[iStateStatPresent].dblProbabilityVariance	= 0;
			}

			if (iStateStatMissing < (INT) dmattributestat.ulStates)
			{
				// p(missing) = 0
				dmattributestat.rgStateStats[iStateStatMissing].Value = DmstatevalueCategorical(dblMissing); 
				dmattributestat.rgStateStats[iStateStatMissing].dblProbability = 0.0;
				dmattributestat.rgStateStats[iStateStatMissing].dblSupport = 0.0;
				dmattributestat.rgStateStats[iStateStatMissing].dblVariance = 0; 
				dmattributestat.rgStateStats[iStateStatMissing].dblProbabilityVariance = 0; 
			}
		}
		else
		{
			dmattributestat.ulStates = 0;
		}

		if (iPrediction == in_ulMaxPredictions)
		{
			// There's no sort order for the attributes in our model, so
			// we'll simply stop once we've reached the maximum.
			*out_pulPredictions = in_ulMaxPredictions;
			return S_OK;
		}
	}

	// Because there's no missing data, all the attributes are equivalent in terms of any
	// sort order that the caller can request. Thus, we need not worry about performing a sort.
	*out_pulPredictions = iPrediction;
	return S_OK;
}


STDMETHODIMP ALGORITHM::GetNodeIDsForCase( 
        /* [in] */ IDMContextServices*	in_pContext,
        /* [in] */ DM_NODE_FLAGS		in_Flags,
        /* [in] */ DM_CaseID			in_CaseID,
        /* [in] */ ULONG				in_ulCaseValues,
        /* [in] */ DM_ATTRIBUTE_VALUE*	in_rgValues,
        /* [in] */ ULONG				in_ulMaxNodeIDs,
        /* [out] */ DMStringPtr**		io_prgNodeIDs,
        /* [out] */ ULONG*				out_pulNodeIDs)
{
	// This is the main function used for the "DM Dimension" slicing and "drill
	// through". The idea is that your model implements a mapping from cases into 
	// "node ids". For "DM Dimension" mode, the mapping results, for each case, 
	// in exactly one node id for each output variable. In non-DM-Dimension mode, 
	// there can be multiple node ids for each output variable.
	
	// The input DM_NODE_FLAGS specifies that you are to implement the DM-dimension
	// mapping by setting the DMNF_DIMENSION_CONTENT_NODE bits.

	// The node ids returned by this function are the unique names
	// (GetNodeIDFromUniqueName) used by IDMAlgorithmNavigation. You must check
	// if *io_prgNodeIDs is NULL, and if so, you must allocate the appropriate
	// array.

	// For our model, we'll build the following DM Dimension hierarchy: for each 
	// target attribute, we'll test whether the attribute has a posterior mean > 0. 
	// Depending on the outcome, we'll return a different leaf node. The naming scheme
	// and implementation is contained in "Hierarchies.h"
	//
	// For the "Drill through", we need to be consistent with the decision-tree
	// viewer, which means that we have a single root node (i.e. there's no splits 
	// in our "decision tree") that corresponds to every case. 

	UINT ciAttributeOutput = (UINT) _lrpmodel._viAttributeOutput.size();

	if (in_ulMaxNodeIDs < ciAttributeOutput)
	{
		return E_FAIL;
	}

	*out_pulNodeIDs = ciAttributeOutput;

	CComPtr<IDMStringHandler>	spidmstringhandler		= NULL;
	CComPtr<IDMMemoryAllocator>	spidmmemoryallocator	= NULL;
	
	HRESULT hr = in_pContext->GetStringHandler(&spidmstringhandler);

	RETURN_ON_FAIL(hr);

	if (*io_prgNodeIDs == NULL)
	{
		// We have to allocate this array. Regardless of the type of
		// hierarchy, we're going to have one node ID per output attribute.
		hr = in_pContext->GetVectorAlloc(&spidmmemoryallocator);

		RETURN_ON_FAIL(hr);

		DMStringPtr*	rgdmstringNodeIDs = NULL;


		hr = spidmmemoryallocator->Alloc(ciAttributeOutput * sizeof(DMStringPtr), (byte**) &rgdmstringNodeIDs);
		RETURN_ON_FAIL(hr);
		
		memset( rgdmstringNodeIDs, 0, ciAttributeOutput * sizeof(DMStringPtr) );


		for (ULONG iiAttribute = 0; iiAttribute < ciAttributeOutput; iiAttribute++)
		{
			hr = spidmstringhandler->CreateNewStringHandle(&rgdmstringNodeIDs[iiAttribute]);

			RETURN_ON_FAIL(hr);
		}

		*io_prgNodeIDs = rgdmstringNodeIDs;
	}

	DMStringPtr*	rgdmstringNodeIDs = *io_prgNodeIDs;
	WCHAR szNodeID[200];

	if (in_Flags & DMNF_DIMENSION_CONTENT_NODE)
	{
		// We'll construct the hierarchy as described, based on the 
		// posterior distribution. As in Predict(), we'll first convert the 
		// case into a dense vector.

		VDBL	vdblCaseDense(*this);
		vdblCaseDense.resize(_lrpmodel._cAttribute, dblMissing);
		for (UINT iAVPair = 0; iAVPair < in_ulCaseValues; iAVPair++)
		{
			DM_ATTRIBUTE_VALUE& dmattributevalue = in_rgValues[iAVPair];
			ULONG iAttribute = dmattributevalue.Attribute;

			// Check if the attribute is "unspecified" and if so, ignore.
			if (iAttribute == DM_UNSPECIFIED)
			{
				continue;
			}
				
			// Note that the translation of the values MUST match that which
			// is done in the CASEPROCESSOR above in InsertCases(). Because we 
			// only have continuous variables, this is a non-issue for us, but 
			// you need to pay attention to this for categorical variables.
			DBL	 dblValue	= ::DblContinuous(dmattributevalue.Value);
			vdblCaseDense[iAttribute]	= dblValue;
		}

		// Get the posterior for each attribute, and set the appropriate name

		LRPHIERARCHY lrphierarchy(true /* bDMDimension*/, ciAttributeOutput);

		for (ULONG iiAttribute = 0; iiAttribute < ciAttributeOutput; iiAttribute++)
		{
			ULONG iAttribute = _lrpmodel._viAttributeOutput[iiAttribute];
			ULONG	iIDRoot = lrphierarchy.IIDRoot(iiAttribute);
			DBL dblPostMean = 0;
			DBL dblPostSD	= 0;
			hr = _lrpmodel.ExtractPosterior(vdblCaseDense, iAttribute, dblPostMean, dblPostSD);
			RETURN_ON_FAIL(hr);

			ULONG iIDLeaf;
			if (dblPostMean > 0)
			{
				iIDLeaf = lrphierarchy.IIDChild(iIDRoot, 0);
			}
			else
			{
				iIDLeaf = lrphierarchy.IIDChild(iIDRoot, 1);
			}

			hr = lrphierarchy.GetSzFromID(szNodeID, iIDLeaf, 199);
			RETURN_ON_FAIL(hr);
			hr = spidmstringhandler->CopyBufferToHandle(rgdmstringNodeIDs[iAttribute], 
													    szNodeID,
														(UINT) wcslen(szNodeID));
			RETURN_ON_FAIL(hr);
		}

		return S_OK;
	}
	else
	{
		// Drill-through mode : use the root node id for each attribute

		LRPHIERARCHY lrphierarchy(false /* bDMDimension*/, ciAttributeOutput);

		for (ULONG iiAttribute = 0; iiAttribute < ciAttributeOutput; iiAttribute++)
		{
			ULONG	iAttribute	= _lrpmodel._viAttributeOutput[iiAttribute];
			ULONG	iIDRoot		= lrphierarchy.IIDRoot(iiAttribute);
			hr = lrphierarchy.GetSzFromID(szNodeID, iIDRoot, 199);
			RETURN_ON_FAIL(hr);
			hr = spidmstringhandler->CopyBufferToHandle(rgdmstringNodeIDs[iiAttribute], 
													    szNodeID,
														(UINT) wcslen(szNodeID));
			RETURN_ON_FAIL(hr);
		}
	}

	return S_OK;
}



STDMETHODIMP ALGORITHM::GetNavigator( 
    /* [in] */	IDMContextServices *in_pContext,
    /* [in] */	BOOL in_fDimension,
    /* [out] */ IDMAlgorithmNavigation **out_ppDAGNav)
{
	// This function returns an algorithm-navigation interface. 
	// Several instances of this interface may be active concurrently 
	// over a single algorithm instance, so you want to implement
	// the navigation with a separate class.
	//
	// Assuming that the IDMAlgorithmMetadata interface is implemented
	// using static member functions, the following check will be 
	// appropriate. Otherwise, you should delete this code.

/*	BOOL bSupportDMDimensions;

	hr = GetSupportsDMDimensions(in_pContext, &bSupportDMDimensions);

	RETURN_ON_FAIL(hr);

	if (in_fDimension && !bSupportDMDimensions)
	{
		// The algorithm has declared that it does not support
		// DM dimensions, yet the caller is attempting to get
		// a navigator appropriate for DM dimensions.

		return E_FAIL;
	}
*/	

	// Create an instance of your navigation class and return the
	// desired interface. For example:
 
	CComObjectDM<NAVIGATOR>*	pNavClass;
	HRESULT hr = CComObjectDM<NAVIGATOR>::CreateInstance(*this, &pNavClass);
	RETURN_ON_FAIL(hr);

	hr = pNavClass->Initialize((in_fDimension == TRUE) ? true : false, this);
	RETURN_ON_FAIL(hr);

	if (pNavClass)
	{
		hr = pNavClass->QueryInterface(
						__uuidof(IDMAlgorithmNavigation), 
						(void**) out_ppDAGNav);
		RETURN_ON_FAIL(hr);
	}
	
	return hr;
}
	
STDMETHODIMP	ALGORITHM::GetSampleCaseSet( 
        /* [in] */ IDMContextServices*	in_pContext,
        /* [in] */ DMString*			in_pstrNodeID,
        /* [out] */ IDMPullCaseSet**	out_ppCaseSet)
{
	// This function is supposed to return a set of cases that are randomly
	// sampled in some way that is "appropriate" for a particular node in the 
	// model. Depending on your model, this may or may not make any sense.

	return S_OK;
}

STDMETHODIMP ALGORITHM::GetTrainingParameterActualValue( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [in] */ ULONG in_iParameter,
    /* [in] */ DMVariantPtr io_pParameterValue)
{ 
	VARIANT	varTmp;
	::VariantInit(&varTmp);
	switch( in_iParameter )
	{
	case 0:
		V_VT(&varTmp)	=	VT_R8;
		V_R8(&varTmp)	=	(DOUBLE)_dblMinDepScore;
		break;
	case 1:
		V_VT(&varTmp)	=	VT_BOOL;
		V_R8(&varTmp)	=	(BOOL)_bDisplayCorrelation;
		break;
	default:
		// Parameter index out of range
		return E_INVALIDARG;
	}

	// Smart pointer for a variant handler
	CComPtr<IDMVariantPtrHandler> spidmvarianthandler;
	HRESULT hr = in_pContext->GetVariantHandler(&spidmvarianthandler);
	RETURN_ON_FAIL(hr);
	return spidmvarianthandler->CopyVariantToHandle(io_pParameterValue, &varTmp);
}


STDMETHODIMP ALGORITHM::HasFeatureSelection( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [out] */ BOOL *out_pHasFS)
{ 
	HRESULT hr = S_OK;
	return hr; 
}


STDMETHODIMP ALGORITHM::GetFeatureSelectedAttributes( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [out] */ ULONG *out_pCount,
    /* [out] */ DM_Attribute **out_prgFeatureSelectedAttributes,
    /* [out] */ DM_ATTRIBUTE_FLAGS **out_prgFSAttributeFlags)
{ 
	HRESULT hr = S_OK;
	return hr; 
}


STDMETHODIMP ALGORITHM::GetAttributeFeatureSelectionFlags( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [in] */ DM_Attribute in_nAttr,
    /* [out] */ DM_ATTRIBUTE_FLAGS *out_pAttFlags)
{ 
	HRESULT hr = S_OK;
	return hr; 
}


STDMETHODIMP ALGORITHM::GetProcessingMemoryEstimate( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [out] */ LONG *out_pEstimate)
{ 
	HRESULT hr = S_OK;
	return hr; 
}


void STDMETHODCALLTYPE ALGORITHM::UpdateMemoryEstimate( 
    /* [in] */ LONG in_lEstimateChange)
{ 
	return; 
}


LONG STDMETHODCALLTYPE ALGORITHM::ReservedMemoryEstimate( void )
{ 
	return 0; 
}


STDMETHODIMP ALGORITHM::RenderPMMLContent( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [in] */ ISAXContentHandler *in_pXMLWriter)
{ 
	HRESULT hr = S_OK;
	return hr; 
}


STDMETHODIMP ALGORITHM::PreInitializeForPMML21Parsing( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [in] */ IDMAttributeSet *in_pAttributeSet)
{ 
	HRESULT hr = S_OK;
	return hr; 
}


STDMETHODIMP ALGORITHM::GetPMMLAlgorithmSAXHandler( 
    /* [in] */ IDMContextServices *in_pContext,
    /* [out] */ ISAXContentHandler **out_ppSAXHandler)
{ 
	HRESULT hr = S_OK;
	return hr; 
}