/*=====================================================================

  File:      ClusterGridViewer.cs
  Summary:   Data Mining Web Controls is a library sample intended 
			 to extend the SQL Server Data Mining user experience 
			 to the Web. The library provides a lightweight version 
			 of the Data Mining model viewers. Using Data Mining 
			 Web Controls, complex mining models can be browsed 
			 from any machine that has Microsoft Internet Explorer 
			 installed.
  Date:      Oct 20, 2003

---------------------------------------------------------------------

  This file is part of the Microsoft SQL Server Code Samples.
  Copyright (C) 2003 Microsoft Corporation.  All rights reserved.

This source code is intended only as a supplement to Microsoft
Development Tools and/or on-line documentation.  See these other
materials for detailed information regarding Microsoft code samples.

THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.

======================================================= */
namespace Microsoft.AnalysisServices.DataMiningHtmlViewers
{
    using System;
    using System.CodeDom;
    using System.Collections;
    using System.ComponentModel;
    using System.ComponentModel.Design.Serialization;
    using System.Configuration;
    using System.Data;
    using System.Globalization;
    using System.IO;
    using System.Security.Permissions;
    using System.Reflection;
    using System.Web;
    using System.Web.UI;
    using System.Web.UI.WebControls;

    /// <summary>
    /// Cluster Discrimination - compares the attributes from two 
    /// different clusters and displays the attribute state 
    /// associated with differences in order of importance.
    /// Cluster Characteristics - characteristics that make up a
    /// specific cluster.
    /// </summary>
    public enum GridViewerMode
    {
        Discrimination,
        Characteristics
    }

    /// <summary>
    /// The DMClusterViewer control can be used to display the 
    /// content of models created using the Microsoft_Clustering 
    /// Data Mining algorithm. Not as rich as the Cluster viewer 
    /// from the Business Intelligence Workbench, the cluster viewer 
    /// only supports Cluster Characteristics and Cluster Discrimination.
    /// </summary>
    [AspNetHostingPermission(SecurityAction.LinkDemand, Level = AspNetHostingPermissionLevel.Minimal)]
    [AspNetHostingPermission(SecurityAction.InheritanceDemand, Level = AspNetHostingPermissionLevel.Minimal)]
    public abstract class ClusterGridViewer : DMHtmlViewer, INamingContainer
    {
        private System.Web.UI.WebControls.Table table;
        private System.Drawing.Color colorBar;
        private GridViewerMode gridMode;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Usage", "CA2211:NonConstantFieldsShouldNotBeVisible")]
        public static String Missing = Properties.Resources.StrMissing;

        protected ClusterGridViewer()
        {
            colorBar = System.Drawing.Color.Navy;
        }

        // Override the Control methods.
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security", "CA2109:ReviewVisibleEventHandlers")]
        protected override void OnLoad(EventArgs e)
        {
            EnsureChildControls();
            InitializeContent();
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Security", "CA2109:ReviewVisibleEventHandlers")]
        protected override void OnInit(EventArgs e)
        {
            EnsureChildControls();
            InitializeContent();
        }

        protected override void CreateChildControls()
        {
            RenderChildControls();
        }

        /// <summary>
        /// Save state as a cumulative array of objects.
        /// </summary>
        /// <returns>Returns an array containing the properties of the control.</returns>
        protected override object SaveViewState()
        {
            object baseState = base.SaveViewState();
            object[] allStates = new object[2];
            allStates[0] = baseState;
            allStates[1] = gridMode;
            return allStates;
        }

        /// <summary>
        /// Set class members to the property values.
        /// </summary>
        /// <param name="savedState">Returns an object containing the state information.</param>
        protected override void LoadViewState(object savedState)
        {
            if (savedState != null)
            {
                object[] myState = (object[])savedState;
                if (myState[0] != null)
                    base.LoadViewState(myState[0]);
                if (myState[1] != null)
                    gridMode = (GridViewerMode)myState[1];
            }
        }

        /// <summary>
        /// Generate jscript code for the Html page.
        /// </summary>
        /// <param name="output">Renders ...</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1725:ParameterNamesShouldMatchBaseDeclaration")]
        protected override void Render(HtmlTextWriter output)
        {
            string tooltipScript = "<Script language='JavaScript'>\n";

            tooltipScript += "var __oDMClustGridTooltip = null;\n";
            tooltipScript += "function DMClustGridHideTooltip()\n";
            tooltipScript += "{\n";
            tooltipScript += "__oDMClustGridTooltip.hide();\n}\n";

            tooltipScript += "function DMClustGridShowTooltip()\n";
            tooltipScript += "{\n";
            tooltipScript += "    if (__oDMClustGridTooltip == null) __oDMClustGridTooltip = window.createPopup();\n";
            tooltipScript += "    DivElement = window.event.srcElement.children[0];\n";
            tooltipScript += "    if( DivElement == null )return;\n";
            tooltipScript += "    __oDMClustGridTooltip.document.body.innerHTML = DivElement.innerHTML;\n";
            tooltipScript += "    var popupBody = __oDMClustGridTooltip.document.body;\n";

            if (gridMode == GridViewerMode.Discrimination)
            {
                tooltipScript += "    __oDMClustGridTooltip.show(0, 0, 400, 0);\n";
            }
            else
            {
                tooltipScript += "    __oDMClustGridTooltip.show(0, 0, 100, 0);\n";
            }

            tooltipScript += "    var realHeight = popupBody.scrollHeight;\n";
            tooltipScript += "    __oDMClustGridTooltip.hide();\n";

            if (gridMode == GridViewerMode.Discrimination)
            {
                tooltipScript += "    __oDMClustGridTooltip.show(60, 10, 400, realHeight, event.srcElement);\n";
            }
            else
            {
                tooltipScript += "    __oDMClustGridTooltip.show(60, 10, 100, realHeight, event.srcElement);\n";
            }

            tooltipScript += "}\n";
            tooltipScript += "</Script>";

            if (!IsDesignTime())
            {
                output.Write(tooltipScript);
                output.Write(tooltipScript);
            }
            base.Render(output);
        }
        // End Override

        /// <summary>
        /// Reinitialize the controls on the page.
        /// </summary>
        protected void ResetContent()
        {
            Controls.Clear();
            RenderChildControls();
            InitializeContent();
        }


        [
        LocCategory("DataMiningLayout"),
        LocDescription("DescGBColor")
        ]
        public System.Drawing.Color BarColor
        {
            get
            {
                return colorBar;
            }
            set
            {
                colorBar = value;
                if (table != null)
                    ResetContent();
            }
        }

        [
        LocCategory("DataMiningLayout"),
        DefaultValue(GridViewerMode.Discrimination),
        LocDescription("DescGridViewerMode")
        ]
        public GridViewerMode ViewerMode
        {
            get
            {
                return gridMode;
            }

            set
            {
                gridMode = value;
                if (table != null)
                    ResetContent();
            }
        }

        protected abstract void RenderChildControls();

        protected abstract void InitializeContent();

        /// <summary>
        /// Define a table and add it to the list of viewable controls. 
        /// </summary>
        protected void CreateTable()
        {
            table = new Table();
            table = new Table();
            table.Width = this.Width;
            table.GridLines = GridLines.Both;
            Controls.Add(table);

            ResetTable();
        }

        /// <summary>
        /// Set the table label text depending on the state 
        /// (gridMode = Discrimination or Characteristics).
        /// </summary>
        /// <param name="stateName">Name of the view.</param>
        protected void SetLeftStateName(string stateName)
        {
            System.Web.UI.WebControls.TableRow row = table.Rows[0];
            if (gridMode == GridViewerMode.Discrimination)
            {
                row.Cells[2].Text = Properties.Resources.StrFavors + stateName;
            }
            else
            {
                row.Cells[2].Text = Properties.Resources.StrFrequency;
            }
        }

        /// <summary>
        /// Set the table label text depending on the state 
        /// (gridMode = Discrimination or Characteristics).
        /// </summary>
        /// <param name="stateName">Name of the view.</param>
        protected void SetRightStateName(string stateName)
        {
            if (gridMode == GridViewerMode.Discrimination)
            {
                System.Web.UI.WebControls.TableRow row = table.Rows[0];
                row.Cells[3].Text = Properties.Resources.StrFavors + stateName;
            }
        }

        /// <summary>
        /// Reinitialize the table to the default settings.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        protected void ResetTable()
        {
            try
            {
                table.Rows.Clear();
                table.Visible = true;

                System.Web.UI.WebControls.TableRow row = new System.Web.UI.WebControls.TableRow();
                System.Web.UI.WebControls.TableHeaderCell hdrCell;

                hdrCell = new TableHeaderCell();
                hdrCell.BackColor = System.Drawing.Color.LightGray;
                hdrCell.Font.Bold = true;
                hdrCell.Text = Properties.Resources.StrAttributes;
                row.Cells.Add(hdrCell);

                hdrCell = new TableHeaderCell();
                hdrCell.BackColor = System.Drawing.Color.LightGray;
                hdrCell.Font.Bold = true;
                hdrCell.Text = Properties.Resources.StrValues;
                row.Cells.Add(hdrCell);

                hdrCell = new TableHeaderCell();
                hdrCell.BackColor = System.Drawing.Color.LightGray;
                hdrCell.Font.Bold = true;
                hdrCell.Text = (gridMode == GridViewerMode.Discrimination) ? Properties.Resources.StrFavorsV1 : Properties.Resources.StrFrequency;
                hdrCell.Width = 100;
                row.Cells.Add(hdrCell);

                if (gridMode == GridViewerMode.Discrimination)
                {
                    hdrCell = new TableHeaderCell();
                    hdrCell.BackColor = System.Drawing.Color.LightGray;
                    hdrCell.Font.Bold = true;
                    hdrCell.Width = 100;
                    hdrCell.Text = Properties.Resources.StrFavorsV2;
                    row.Cells.Add(hdrCell);
                }

                table.Rows.Add(row);
            }
            catch (Exception /*ex*/)
            {
                // Add code to show the error message.
            }
        }

        /// <summary>
        /// Add a row to the table.
        /// </summary>
        /// <param name="leftCell">Text for the left cell.</param>
        /// <param name="rightCell">Text for the right cell.</param>
        /// <param name="cellWidth">Width of the cells.</param>
        protected void AddTableRow(string leftCell, string rightCell, double cellWidth)
        {
            AddTableRow(leftCell, rightCell, cellWidth, "");
        }

        /// <summary>
        /// Add a row to the table.
        /// </summary>
        /// <param name="leftCell">Text for the left cell.</param>
        /// <param name="rightCell">Text for the right cell.</param>
        /// <param name="cellWidth">Width of the cells.</param>
        /// <param name="toolTip">Text for a tooltip on the table.</param>
        protected void AddTableRow(string leftCell, string rightCell, double cellWidth, string toolTip)
        {
            try
            {
				cellWidth = -cellWidth;

                System.Web.UI.WebControls.TableRow row = new System.Web.UI.WebControls.TableRow();
                System.Web.UI.WebControls.TableCell cell;

                cell = new TableCell();
                cell.Text = leftCell;
                row.Cells.Add(cell);

                cell = new TableCell();
                cell.Text = rightCell;
                row.Cells.Add(cell);


                if (gridMode == GridViewerMode.Characteristics)
                {
                    if (cellWidth < 0)
                        cellWidth = -cellWidth;
                }

                int widthAbs;
                int widthInt;

                cell = new TableCell();

                if (cellWidth < 0)
                {
                    // Favors left
                    widthInt = System.Convert.ToInt32(-cellWidth);
                    widthAbs = widthInt;
                    cell.HorizontalAlign = HorizontalAlign.Right;
                }
                else
                {
                    // Favors right
                    if (gridMode == GridViewerMode.Discrimination)
                    {
                        cell.Text = "&nbsp;";
                        row.Cells.Add(cell);
                    }

                    cell = new TableCell();
                    widthInt = System.Convert.ToInt32(cellWidth);
                    widthAbs = widthInt;
                    cell.HorizontalAlign = HorizontalAlign.Left;
                }

                if (toolTip.Length == 0)
                {
                    toolTip = widthAbs.ToString(CultureInfo.CurrentCulture);
                }

                string cellText;
				cellText = "<DIV onmouseover=\"DMClustGridShowTooltip()\" onmouseout=\"DMClustGridHideTooltip()\" STYLE=\"font-size:0%; border:none; height:10px; left:5px; ";
                cellText += "width:" + widthAbs.ToString(CultureInfo.CurrentCulture) + "px; ";
                cellText += "background-color:" + System.Drawing.ColorTranslator.ToHtml(colorBar) + "; ";
                cellText += "\" >";
                cellText += "<DIV STYLE=\"visibility:hidden; overflow:hidden; position:absolute; left:0; right:0; width:1; height:1\">" + toolTip + "</DIV>";
                cellText += "</DIV>";
                cell.Text = cellText;
                row.Cells.Add(cell);

                if (cellWidth < 0)
                {
                    cell = new TableCell();
                    cell.Text = "&nbsp;";
                    row.Cells.Add(cell);
                }

                table.Rows.Add(row);
            }
            catch (Exception /*ex*/)
            {
                // Add code to show the error message.
            }
        }
    }
}
